var ffi = require('ffi-napi');
const EventEmitter = require('events');
var timer = require("node:timers/promises");
var os = require("os");
var folder = process.cwd();
class Tdlib {

    constructor(path_tdlib = "", option_tdlib = {}) {

        if (!option_tdlib) {
            option_tdlib = {};
        } else if (typeof option_tdlib != "object") {
            option_tdlib = {};
        }
        this.is_android = (os.platform() == "android");
        var method_tdlib = {
            'td_create_client_id': ["int", []],
            'td_send': ["void", ["int", "char"]],
            'td_receive': ["char", ["double"]],
            'td_execute': ["char", ["char"]],
            'td_json_client_create': ['pointer', []],
            'td_json_client_send': ['void', ['pointer', 'string']],
            'td_json_client_receive': ['string', ['pointer', 'double']],
            'td_json_client_execute': ['string', ['pointer', 'string']],
            'td_json_client_destroy': ['void', ['pointer']],
            'td_set_log_file_path': ['int', ['string']],
            'td_set_log_verbosity_level': ['void', ['int']],
            'td_set_log_fatal_error_callback': ['void', ['pointer']]
        };
        if (this.is_android) {
            method_tdlib = {
                '_td_create_client_id': ["int", []],
                '_td_send': ["void", ["int", "char"]],
                '_td_receive': ["char", ["double"]],
                '_td_execute': ["char", ["char"]],
                '_td_json_client_create': ['pointer', []],
                '_td_json_client_send': ['void', ['pointer', 'string']],
                '_td_json_client_receive': ['string', ['pointer', 'double']],
                '_td_json_client_execute': ['string', ['pointer', 'string']],
                '_td_json_client_destroy': ['void', ['pointer']],
                '_td_set_log_file_path': ['int', ['string']],
                '_td_set_log_verbosity_level': ['void', ['int']],
                '_td_set_log_fatal_error_callback': ['void', ['pointer']]
            };
        }
        this.libm = ffi.Library(path_tdlib, method_tdlib);
        this.optionTdlibDefault = {
            "@type": 'setTdlibParameters',
            'parameters': {
                '@type': 'tdlibParameters',
                'api_id': 0,
                'api_hash': '',
                'database_directory': `${folder}/./db_client`,
                'files_directory': `${folder}/./db_client`,
                "use_file_database": false,
                "use_chat_info_database": false,
                "use_message_database": false,
                "use_secret_chats": false,
                'enable_storage_optimizer': false,
                'system_language_code': 'en',
                'new_verbosity_level': 0,
                'application_version': 'v1',
                'device_model': 'Telegram Client Hexaminate',
                'system_version': "Linux",
                "database_key": "",
                "start": true,
                ...option_tdlib
            }
        };
        this.event_emitter = new EventEmitter();
        this.event_emitter.setMaxListeners(0);
        this.libm[`${this.is_android ? "_" : ""}td_set_log_verbosity_level`](this.optionTdlibDefault.parameters.new_verbosity_level);
        this.client = this.libm[`${this.is_android ? "_" : ""}td_json_client_create`]();
    }

    client_destroy(parameters = {}) {
        return this.libm[`${this.is_android ? "_" : ""}td_json_client_destroy`](this.client, this.json(parameters));
    }

    uuId(bytesNumber = 10) {
        var text = "1234567890qwertyuiopasdfghjklzxcvbnm";
        var texts = [...text];
        var message = "";
        while (true) {
            var data = texts[Math.floor(Math.random() * texts.length)];
            message += data;
            if (message.length == bytesNumber) {
                return message;
            }
        }
    }

    async client_send(parameters = {}, is_without_result = false) {
        var unique_id = this.uuId();
        parameters["@extra"] = unique_id;
        this.libm[`${this.is_android ? "_" : ""}td_json_client_send`](this.client, this.json(parameters));
        if (is_without_result) {
            return null;
        }
        var json_result = false;
        this.on("update", async function (update) {
            if (update["@extra"] == unique_id) {
                delete update["@extra"];
                json_result = update;
            }
        });
        while (true) {
            await timer.setTimeout(1);
            if (json_result) {
                if (json_result["@type"] == "error") {
                    throw json_result;
                }
                return json_result;
            }
        }
    }

    client_execute(parameters = {}) {
        var json_result = JSON.parse(this.libm[`${this.is_android ? "_" : ""}td_json_client_execute`](this.client, this.json(parameters)));
        if (json_result["@type"] == "error") {
            throw json_result;
        }
        return json_result;
    }


    invokeSync(method, parameters = {}) {
        var is_without_result = (parameters["is_without_result"] ? true : false);
        if (!parameters["@type"]) {
            parameters["@type"] = method;
        }
        delete parameters["is_without_result"];
        return this.client_execute(parameters, is_without_result);
    }

    async invoke(method, parameters = {}) {
        var is_without_result = (parameters["is_without_result"] ? true : false);
        if (!parameters["@type"]) {
            parameters["@type"] = method;
        }
        delete parameters["is_without_result"];
        if (parameters["sync"]) {
            return this.client_execute(parameters, is_without_result);
        } else {
            return await this.client_send(parameters);
        }
    }

    client_receive(timeout = 1.0) {
        var result = this.libm[`${this.is_android ? "_" : ""}td_json_client_receive`](this.client, timeout);
        if (result != null) {
            return JSON.parse(result);
        }
        return result;
    }

    async requestApi(method, parameters = {}) {
        return await this.invoke(method, parameters);
    }

    async request(method, parameters = {}) {

        if (typeof parameters["chat_id"] == "string" && parameters["chat_id"] && RegExp("@?[a-z0-9_]+", "i").exec(parameters["chat_id"])) {
            var search_public_chat = await this.invoke("searchPublicChat", { "username": parameters["chat_id"] });
            if (search_public_chat["@type"] == "chat") {
                parameters["chat_id"] = search_public_chat["id"];
            }
        }
        if (typeof parameters["user_id"] == "string" && parameters["user_id"] && RegExp("@?[a-z0-9_]+", "i").exec(parameters["user_id"])) {
            var search_public_chat = await this.invoke("searchPublicChat", { "username": parameters["chat_id"] });
            if (search_public_chat["@type"] == "chat") {
                parameters["user_id"] = search_public_chat["id"];
            }
        }

        if (RegExp("^getChat$", "i").exec(method)) {
            return await this.getChat(parameters["chat_id"], parameters["is_api_update"]);
        }
        if (RegExp("^getUser$", "i").exec(method)) {
            return await this.getUser(parameters["chat_id"], parameters["is_api_update"]);
        }
        if (RegExp(`^getChatAdministrators$`, "i").exec(method)) {

            var json_respond = {
                "status_bool": true,
                "status_code": 200,
                "result": {}
            };
            var data = {
                '@type': "getChatAdministrators",
                "chat_id": parameters["chat_id"]
            };
            var getChatAdministrators = await this.invoke(method, data);
            if (typeof getChatAdministrators != "object") {
                json_respond["status_bool"] = false;
                json_respond["status_code"] = 400;
                return json_respond;
            }
            if (getChatAdministrators["@type"] == "chatAdministrators") {
                if (typeof getChatAdministrators["administrators"] != "object") {
                    json_respond["status_bool"] = false;
                    json_respond["status_code"] = 404;
                    return json_respond;
                }
                var array_list_admin = [];
                for (var i = 0; i < getChatAdministrators["administrators"].length; i++) {
                    var loop_data = getChatAdministrators["administrators"][i];
                    if (typeof loop_data == "object") {
                        if (loop_data["@type"] == "chatAdministrator") {
                            var json_data = {
                                "user": {
                                    "id": loop_data["user_id"]
                                }
                            };
                            try {
                                try {
                                    var getChat = await this.getUser(loop_data["user_id"]);
                                    json_data["user"] = getChat["result"];
                                } catch (e) {

                                }
                                json_data["is_owner"] = loop_data["is_owner"];
                                json_data["status"] = loop_data["is_owner"] ? "creator" : "administrator";
                                json_data["custom_title"] = loop_data["custom_title"];
                                array_list_admin.push(json_data);
                            } catch (e) {

                            }
                        }
                    }
                }
                json_respond["result"] = array_list_admin;
                return json_respond;
            }
        }

        if (RegExp("^deleteMessages$", "i").exec(method)) {
            parameters["message_ids"] = parameters["message_id"].constructor === Array ? parameters["message_id"] : [parameters["message_id"]]
            parameters["revoke"] = (parameters["revoke"]) ? true : false;
            return await this.invoke(method, parameters);
        }
        if (RegExp("^getChatMember$", "i").exec(method)) {
            var option = {
                '@type': "getChatMember",
                "chat_id": parameters["chat_id"],
                "member_id": {
                    "@type": "messageSenderUser",
                    "user_id": parameters["user_id"]
                }
            };
            var getChatMember = await this.invoke(method, option);
            var json = {
                "user": {
                    "id": getChatMember["member_id"]["user_id"]
                }
            };
            try {
                var getUser = await this.request("getUser", { "chat_id": getChatMember["member_id"]["user_id"] });
                json["user"] = getUser["result"];
            } catch (e) {
            }
            json["join_date"] = getChatMember["joined_chat_date"] ?? 0;
            var status = getChatMember["status"] ?? {};
            json["status"] = String(status["@type"]).toLocaleLowerCase().replace(/(chatmemberstatus)/ig, "");
            json["custom_title"] = status["custom_title"] ?? "";
            json["can_be_edited"] = status["can_be_edited"] ?? false;
            json["can_manage_chat"] = status["can_manage_chat"] ?? false;
            json["can_change_info"] = status["can_change_info"] ?? false;
            json["can_post_messages"] = status["can_post_messages"] ?? false;
            json["can_edit_messages"] = status["can_edit_messages"] ?? false;
            json["can_delete_messages"] = status["can_delete_messages"] ?? false;
            json["can_invite_users"] = status["can_invite_users"] ?? false;
            json["can_restrict_members"] = status["can_restrict_members"] ?? false;
            json["can_pin_messages"] = status["can_pin_messages"] ?? false;
            json["can_promote_members"] = status["can_promote_members"] ?? false;
            json["can_manage_voice_chats"] = status["can_manage_voice_chats"] ?? false;
            json["is_anonymous"] = status["is_anonymous"] ?? false;
            return { "ok": true, "result": json };
        }
        if (RegExp("^(sendMessage|sendPhoto|sendVideo|sendAudio|sendVoice|sendDocument|sendSticker|sendAnimation|editMessageText|copyMessage|forwardMessage)$", "i").exec(method)) {
            var result_request = { "ok": false, "result": {} };
            result_request = await this.invoke(method, this.makeParametersApi({
                "@type": method,
                ...parameters
            }));
            if (!parameters["as_api"]) {
                return result_request;
            }
            var result = {};
            var jsonMessage = this;
            this.on("update", async function (update) {
                if (update["@type"] == "updateMessageSendSucceeded") {
                    if (update["old_message_id"] == result_request["id"]) {
                        var json_message = await jsonMessage.jsonMessage(update["message"]);
                        if (json_message["ok"]) {
                            json_message["result"]["@type"] = "updateNewMessage";
                            result = json_message["result"];
                        } else {
                            json_message["result"]["@type"] = "error";
                            result = json_message["result"];
                        }

                    }
                }
            });


            while (true) {
                await timer.setTimeout(1);
                if (typeof (result["@type"]) == "string") {
                    if (result["@type"] == "error") {
                        throw result;
                    }
                    delete result["@type"];
                    return { "ok": true, "result": result };
                }
            }

        }


        return await this.invoke(method, this.makeParametersApi({
            "@type": method,
            ...parameters
        }));
    }

    async getMessage(chat_id, message_id, is_detail = false, is_super_detail = false, from_data = null, chat_data = null) {
        var get_message = await this.invoke("getMessage", {
            "chat_id": chat_id,
            "message_id": message_id,
        });
        return await this.jsonMessage(get_message, is_detail, is_super_detail, from_data, chat_data);
    }
    async jsonMessage(update, is_detail = false, is_super_detail = false, from_data = null, chat_data = null) {
        var json_result = {};
        var json = {};

        if (update["@type"] == "message") {
            var json = {};
            var chat_json = {
                "id": update["chat_id"],
                "first_name": "",
                "title": "",
                "type": "",
                "detail": {},
                "last_message": {}
            };
            if (update["is_channel_post"] ?? false) {
                chat_json["type"] = "channel";
                chat_json["title"] = "";
            } else {
                if (RegExp("^-100", "i").exec(update["chat_id"])) {
                    chat_json["type"] = "supergroup";
                    chat_json["title"] = "";
                } else if (RegExp("^-", "i").exec(update["chat_id"])) {
                    chat_json["type"] = "group";
                    chat_json["title"] = "";
                } else {
                    chat_json["type"] = "private";
                }
            }

            if (!is_super_detail) {
                if (chat_json["type"] != "private") {
                    delete chat_json["first_name"];
                } else {
                    delete chat_json["title"];
                }
            }

            var is_chat_not_same = true;
            if (is_super_detail) {
                try {
                    if (chat_data != null) {
                        if (update["chat_id"] == chat_data["id"]) {
                            is_chat_not_same = false;
                            chat_json = chat_data;
                        }
                    }
                } catch (e) { }
                if (is_chat_not_same) {
                    try {
                        var chatResult = await this.getChat(update["chat_id"]);
                        if (chatResult["ok"]) {
                            chat_json = chatResult["result"];
                        }
                    } catch (e) { }
                }
            }

            json["is_outgoing"] = update["is_outgoing"] ?? false;
            json["is_pinned"] = update["is_pinned"] ?? false;
            if (typeof (update["sender_id"]) == "object") {
                var from_json = {
                    "id": 0,
                    "first_name": "",
                    "title": "",
                    "type": "",
                    "detail": {},
                    "last_message": {}
                };
                if (update["sender_id"]["user_id"] != null) {
                    from_json["id"] = update["sender_id"]["user_id"];
                    if (update["chat_id"] == from_json["id"]) {
                        from_json["type"] = chat_json["type"];
                    } else if (RegExp("^-", "i").exec(from_json["chat_id"])) {
                        from_json["type"] = "group";
                    } else {
                        from_json["type"] = "private";
                    }

                    if (is_super_detail) {
                        if (update["chat_id"] == update["sender_id"]["user_id"]) {
                            from_json = chat_json;
                        } else {
                            var is_from_not_same = true;
                            if (from_data != null) {
                                if (update["chat_id"] == from_data["id"]) {
                                    is_from_not_same = false;
                                    from_json = from_data;
                                }
                            }
                            if (is_from_not_same) {
                                try {
                                    var fromResult = await this.getUser(update["sender_id"]["user_id"]);
                                    if (fromResult["ok"]) {
                                        from_json = fromResult["result"];
                                    }
                                } catch (e) { }
                            }
                        }
                    }
                }

                if (update["sender_id"]["chat_id"] != null) {
                    from_json["id"] = update["sender_id"]["chat_id"];
                    if (update["chat_id"] == from_json["id"]) {
                        from_json["type"] = chat_json["type"];
                    } else if (RegExp("^-", "i").exec(from_json["chat_id"])) {
                        from_json["type"] = "group";
                    } else {
                        from_json["type"] = "private";
                    }

                    if (is_super_detail) {
                        if (update["chat_id"] == update["sender_id"]["chat_id"]) {
                            from_json = chat_json;
                        } else {
                            var is_from_not_same = true;
                            if (from_data != null) {
                                if (update["chat_id"] == from_data["id"]) {
                                    is_from_not_same = false;
                                    from_json = from_data;
                                }
                            }
                            if (is_from_not_same) {
                                try {
                                    var fromResult = await this.getChat(update["sender_id"]["chat_id"]);
                                    if (fromResult["ok"]) {
                                        from_json = fromResult["result"];
                                    }
                                } catch (e) { }
                            }
                        }
                    }
                }

                if (!is_super_detail) {
                    if (from_json["type"] != "private") {
                        delete chat_json["first_name"];
                    } else {
                        delete chat_json["title"];
                    }
                }
                json["from"] = from_json;
            }


            json["chat"] = chat_json;
            json["date"] = update["date"];
            json["message_id"] = update["id"];

            if (chat_json["type"] == "channel") {
                if (update["author_signature"]) {
                    json["author_signature"] = update["author_signature"];
                }
            }

            if (typeof (update["forward_info"]) == "object") {
                var forward_info = update["forward_info"];
                if (typeof (forward_info["origin"]) == "object") {
                    if (forward_info["origin"]["@type"] == "messageForwardOriginChannel") {
                        var forward_json = {
                            "id": forward_info["origin"]["chat_id"],
                            "first_name": "",
                            "title": "",
                            "type": "",
                            "detail": {},
                            "last_message": {}
                        };
                        try {
                            var getchat_forward = await this.getChat(forward_json["id"]);
                            if (getchat_forward["ok"]) {
                                forward_json = getchat_forward["result"];
                            }
                        } catch (e) { }
                        json["forward_from_chat"] = forward_json;
                        json["forward_from_message_id"] = forward_info["origin"]["message_id"] ?? 0;
                        json["forward_from_author_signature"] = forward_info["origin"]["author_signature"] ?? "";
                    }
                    if (forward_info["origin"]["@type"] == "messageForwardOriginUser") {
                        var forward_json = {
                            "id": forward_info["origin"]["sender_user_id"],
                            "first_name": "",
                            "title": "",
                            "type": "",
                            "detail": {},
                            "last_message": {}
                        };
                        try {
                            var getuser_forward = await this.getUser(forward_json["id"]);
                            if (getuser_forward["ok"]) {
                                forward_json = getuser_forward["result"];
                            }
                        } catch (e) { }
                        json["forward_from_chat"] = forward_json;
                    }
                }
                json["forward_date"] = forward_info["date"] ?? 0;
            }
            try {
                update["reply_to_message_id"] ??= 0;
                update["reply_in_chat_id"] ??= 0;
            } catch (e) {

            }
            if (update["reply_to_message_id"] != 0 && update["reply_in_chat_id"] != 0) {
                try {
                    var get_message = await this.getMessage(update["reply_in_chat_id"], update["reply_to_message_id"], false, is_super_detail, from_json, chat_json);
                    if (get_message["ok"]) {
                        json["reply_to_message"] = get_message["result"];
                    }
                } catch (e) { }
            }

            if (typeof (update["content"]) == "object") {
                var old_entities = [];

                if (update["content"]["@type"] == "messageText") {
                    json["type_content"] = "text";
                    if (typeof (update["content"]["text"]) == "object") {
                        if (update["content"]["text"]["@type"] == "formattedText") {
                            json["text"] = update["content"]["text"]["text"];
                            old_entities = update["content"]["text"]["entities"];
                        }
                    }
                }

                if (update["content"]["@type"] == "messagePhoto") {
                    json["type_content"] = "photo";
                    if (typeof (update["content"]["photo"]) == "object") {
                        if (update["content"]["photo"]["@type"] == "photo") {
                            var size_photo = [];
                            var photo = update["content"]["photo"]["sizes"];
                            for (var i = 0; i < photo.length; i++) {
                                var photo_json = photo[i];
                                var json_photo = {};
                                if (photo_json["photo"]["remote"]["@type"] == "remoteFile") {
                                    json_photo["file_id"] = photo_json["photo"]["remote"]["id"];
                                }
                                if (photo_json["photo"]["remote"]["unique_id"] != null) {
                                    json_photo["file_unique_id"] = photo_json["photo"]["remote"]["unique_id"];
                                }
                                json_photo["file_size"] = photo_json["photo"]["size"];
                                json_photo["width"] = photo_json["width"];
                                json_photo["height"] = photo_json["height"];
                                size_photo.push(json_photo);
                            }
                            json["photo"] = size_photo;
                        }
                    }
                }

                if (update["content"]["@type"] == "messageVideo") {
                    json["type_content"] = "video";
                    if (typeof (update["content"]["video"]) == "object") {
                        if (update["content"]["video"]["@type"] == "video") {
                            var json_video = {};
                            var content_video = update["content"]["video"];
                            json_video["duration"] = content_video["duration"];
                            json_video["height"] = content_video["height"];
                            json_video["file_name"] = content_video["file_name"];
                            json_video["mime_type"] = content_video["mime_type"];
                            try {
                                if (update["content"]["video"]["thumbnail"] != null && String(update["content"]["video"]["thumbnail"]["@type"]).toLocaleLowerCase() == "thumbnail") {
                                    var content_thumb = content_video["thumbnail"];
                                    var json_thumb = {};
                                    json_video["thumb"] = json_thumb;
                                    json_thumb["file_id"] = content_thumb["file"]["remote"]["id"];
                                    json_thumb["file_unique_id"] = content_thumb["file"]["remote"]["unique_id"];
                                    json_thumb["file_size"] = content_thumb["file"]["size"];
                                    json_thumb["width"] = content_thumb["width"];
                                    json_thumb["height"] = content_thumb["height"];
                                }
                            } catch (e) { }
                            json_video["file_id"] = content_video["video"]["remote"]["id"];
                            json_video["file_size"] = content_video["video"]["size"];
                            json["video"] = json_video;
                        }
                    }
                }

                if (update["content"]["@type"] == "messageAudio") {
                    var type_content = "audio";
                    json["type_content"] = "audio";
                    if (typeof (update["content"]["audio"]) == "object") {
                        if (update["content"]["audio"]["@type"] == "audio") {
                            var json_content = {};
                            var content_update = update["content"][type_content];
                            json_content["duration"] = content_update["duration"];
                            json_content["title"] = content_update["title"];
                            json_content["performer"] = content_update["performer"];
                            json_content["file_name"] = content_update["file_name"];
                            json_content["mime_type"] = content_update["mime_type"];
                            json_content["file_id"] = content_update[type_content]["remote"]["id"];
                            json_content["unique_id"] = content_update[type_content]["remote"]["unique_id"];
                            json_content["file_size"] = content_update[type_content]["size"];
                            json[type_content] = json_content;
                        }
                    }
                }

                if (update["content"]["@type"] == "messageAnimation") {
                    var type_content = "animation";
                    json["type_content"] = "animation";
                    if (typeof (update["content"]["animation"]) == "object") {
                        if (update["content"]["animation"]["@type"] == "animation") {
                            var json_content = {};
                            var content_update = update["content"][type_content];
                            json_content["duration"] = content_update["duration"];
                            json_content["width"] = content_update["width"];
                            json_content["height"] = content_update["height"];
                            json_content["file_name"] = content_update["file_name"];
                            json_content["mime_type"] = content_update["mime_type"];
                            json_content["mime_type"] = content_update["mime_type"];
                            json_content["has_stickers"] = content_update["has_stickers"];

                            try {
                                if (update["content"][type_content]["thumbnail"] != null && String(update["content"][type_content]["thumbnail"]["@type"]).toLocaleLowerCase() == "thumbnail") {
                                    var content_thumb = content_update["thumbnail"];
                                    var json_thumb = {};
                                    json_thumb["file_id"] = content_thumb["file"]["remote"]["id"];
                                    json_thumb["file_unique_id"] = content_thumb["file"]["remote"]["unique_id"];
                                    json_thumb["file_size"] = content_thumb["file"]["size"];
                                    json_thumb["width"] = content_thumb["width"];
                                    json_thumb["height"] = content_thumb["height"];
                                    json_content["thumb"] = json_thumb;
                                }
                            } catch (e) { }
                            json_content["file_id"] = content_update[type_content]["remote"]["id"];
                            json_content["unique_id"] = content_update[type_content]["remote"]["unique_id"];
                            json_content["file_size"] = content_update[type_content]["size"];
                            json[type_content] = json_content;
                        }
                    }
                }

                if (update["content"]["@type"] == "messageContact") {
                    var type_content = "contact";
                    json["type_content"] = type_content;
                    if (typeof (update["content"][type_content]) == "object") {
                        if (update["content"][type_content]["@type"] == type_content) {
                            var json_content = {};
                            var content_update = update["content"][type_content];
                            json_content["phone_number"] = content_update["phone_number"];
                            json_content["first_name"] = content_update["first_name"];
                            json_content["last_name"] = content_update["last_name"];
                            json_content["vcard"] = content_update["vcard"];
                            json_content["user_id"] = content_update["user_id"];
                            json[type_content] = json_content;
                        }
                    }
                }

                if (update["content"]["@type"] == "messagePoll") {
                    var type_content = "poll";
                    json["type_content"] = type_content;
                    if (typeof (update["content"][type_content]) == "object") {
                        if (update["content"][type_content]["@type"] == type_content) {
                            var json_content = {};
                            var content_update = update["content"][type_content];
                            json_content["id"] = content_update["id"];
                            json_content["question"] = content_update["question"];
                            json_content["options"] = content_update["options"];
                            json_content["total_voter_count"] = content_update["total_voter_count"];
                            json_content["recent_voter_user_ids"] = content_update["recent_voter_user_ids"];
                            json_content["is_anonymous"] = content_update["is_anonymous"];
                            json_content["type"] = content_update["type"];
                            json_content["open_period"] = content_update["open_period"];
                            json_content["close_date"] = content_update["close_date"];
                            json_content["is_closed"] = content_update["is_closed"];
                            json[type_content] = json_content;
                        }
                    }
                }

                if (update["content"]["@type"] == "messageDocument") {
                    var type_content = "document";
                    json["type_content"] = type_content;
                    if (typeof (update["content"][type_content]) == "object") {
                        if (update["content"][type_content]["@type"] == type_content) {
                            var json_content = {};
                            var content_update = update["content"][type_content];
                            json_content["file_name"] = content_update["file_name"];
                            json_content["mime_type"] = content_update["mime_type"];
                            json_content["file_id"] = content_update[type_content]["remote"]["id"];
                            json_content["unique_id"] = content_update[type_content]["remote"]["unique_id"];
                            json_content["file_size"] = content_update[type_content]["size"];
                            json[type_content] = json_content;
                        }
                    }
                }

                if (update["content"]["@type"] == "messageSticker") {
                    var type_content = "sticker";
                    json["type_content"] = type_content;
                    if (typeof (update["content"][type_content]) == "object") {
                        if (update["content"][type_content]["@type"] == type_content) {
                            var json_content = {};
                            var content_update = update["content"][type_content];
                            json_content["set_id"] = content_update["set_id"];
                            json_content["width"] = content_update["width"];
                            json_content["height"] = content_update["height"];
                            json_content["emoji"] = content_update["emoji"];
                            json_content["is_animated"] = content_update["is_animated"];
                            json_content["is_mask"] = content_update["is_mask"];

                            try {
                                if (update["content"][type_content]["thumbnail"] != null && String(update["content"][type_content]["thumbnail"]["@type"]).toLocaleLowerCase() == "thumbnail") {
                                    var content_thumb = content_update["thumbnail"];
                                    var json_thumb = {};
                                    json_thumb["file_id"] = content_thumb["file"]["remote"]["id"];
                                    json_thumb["file_unique_id"] = content_thumb["file"]["remote"]["unique_id"];
                                    json_thumb["file_size"] = content_thumb["file"]["size"];
                                    json_thumb["width"] = content_thumb["width"];
                                    json_thumb["height"] = content_thumb["height"];
                                    json_content["thumb"] = json_thumb;
                                }
                            } catch (e) { }
                            json_content["file_id"] = content_update[type_content]["remote"]["id"];
                            json_content["unique_id"] = content_update[type_content]["remote"]["unique_id"];
                            json_content["file_size"] = content_update[type_content]["size"];
                            json[type_content] = json_content;
                        }
                    }
                }

                if (update["content"]["@type"] == "messageVoiceNote") {
                    var type_content = "voice_note";
                    json["type_content"] = type_content;
                    if (typeof (update["content"][type_content]) == "object") {
                        if (update["content"][type_content]["@type"] == "voiceNote") {
                            var json_content = {};
                            var content_update = update["content"][type_content];

                            json_content["duration"] = content_update["duration"];
                            json_content["waveform"] = content_update["waveform"];
                            json_content["mime_type"] = content_update["mime_type"];

                            json_content["file_id"] = content_update["voice"]["remote"]["id"];
                            json_content["unique_id"] = content_update["voice"]["remote"]["unique_id"];
                            json_content["file_size"] = content_update["voice"]["size"];
                            json["voice"] = json_content;
                        }
                    }
                }
                if (update["content"]["@type"] == "messageChatJoinByLink") {
                    json["type_content"] = "new_member";
                    var new_member_from = json["from"];
                    try {
                        delete new_member_from["detail"];
                    } catch (e) { }
                    json["new_members"] = [new_member_from];
                }
                if (update["content"]["@type"] == "messageChatAddMembers") {
                    json["type_content"] = "new_member";
                    var new_members = [];
                    if (is_super_detail) {
                        for (var i = 0; i < update["content"]["member_user_ids"].length; i++) {
                            var loop_data = update["content"]["member_user_ids"][i];
                            try {
                                var result_user = await this.getUser(loop_data);
                                try {
                                    delete result_user["detail"];
                                } catch (e) { }
                                new_members.push(result_user["result"]);
                            } catch (e) {
                                new_members.push({
                                    "id": loop_data,
                                    "is_bot": false,
                                    "first_name": "",
                                    "last_name": "",
                                    "type": "private"
                                });
                            }
                        }
                    } else {
                        new_members = update["content"]["member_user_ids"];
                    }
                    json["new_members"] = new_members;
                }

                if (update["content"]["@type"] == "messageChatDeleteMember") {
                    json["type_content"] = "left_member";
                    var left_member = [];
                    if (is_super_detail) {
                        try {
                            var result_user = await this.getUser(update["content"]["user_id"]);
                            try {
                                delete result_user["detail"];
                            } catch (e) { }
                            left_member.push(result_user["result"]);
                        } catch (e) {
                            left_member.push({
                                "id": update["content"]["user_id"],
                                "is_bot": false,
                                "first_name": "",
                                "last_name": "",
                                "type": "private"
                            });
                        }
                    } else {
                        left_member.push({
                            "id": update["content"]["user_id"],
                            "is_bot": false,
                            "first_name": "",
                            "last_name": "",
                            "type": "private"
                        });
                    }
                    json["left_member"] = left_member;
                }

                // caption
                if (typeof (update["content"]["caption"]) == "object") {
                    if (update["content"]["caption"]["@type"] == "formattedText") {
                        if (update["content"]["caption"]["text"]) {
                            json["caption"] = update["content"]["caption"]["text"];
                        }
                        old_entities = update["content"]["caption"]["entities"];
                    }
                }

                var new_entities = [];
                for (var i = 0; i < old_entities.length; i++) {
                    var data_entities = old_entities[i];
                    try {
                        var json_entities = {};
                        json_entities["offset"] = data_entities["offset"];
                        json_entities["length"] = data_entities["length"];
                        if (data_entities["type"]["@type"] != null) {
                            var type_entities = String(data_entities["type"]["@type"]).toLocaleLowerCase().replace(RegExp("textEntityType", "i"), "").replace(RegExp("textUrl", "i"), "text_link").replace(RegExp("bot_command", "i"), "bot_command").replace(RegExp("mentionname", "i"), "text_mention");
                            json_entities["type"] = type_entities;
                            if (data_entities["type"]["url"] != null) {
                                json_entities["url"] = data_entities["type"]["url"];
                            }
                            if (type_entities == "text_mention" && data_entities["type"]["user_id"] != null) {
                                var entitiesUserId = data_entities["type"]["user_id"];
                                var fromJson = { "id": entitiesUserId };
                                try {
                                    var fromResult = await this.getChat(update["sender_id"]["user_id"]);
                                    if (fromResult["ok"]) {
                                        fromJson = fromResult["result"];
                                    }
                                } catch (e) { }
                                json_entities["user"] = fromJson;
                            }
                        }
                        new_entities.push(json_entities);
                    } catch (e) { }
                }
                json["entities"] = new_entities;
            }
            if (is_detail) {
                if (is_super_detail) {
                    if (json["chat"]["type"] != null) {
                        if (json["chat"]["type"] == "channel") {
                            return {
                                "ok": true,
                                "result": { "update_channel_post": json }
                            };
                        } else {
                            return {
                                "ok": true,
                                "result": { "update_message": json }
                            };
                        }
                    }
                } else {
                    if (json["chat"]["type"] == "channel") {
                        return {
                            "ok": true,
                            "result": { "update_channel_post": json }
                        };
                    } else {
                        return {
                            "ok": true,
                            "result": { "update_message": json }
                        };
                    }
                }
            }
            return { "ok": true, "result": json };

        }

        if (update["@type"] == "updateNewCallbackQuery") {
            var json = {};
            var from = { "id": update["sender_user_id"] };
            var chat = { "id": update["chat_id"] };
            json["id"] = update["id"];

            if (is_super_detail) {
                try {
                    var fromResult = await this.getChat(chat["id"]);
                    if (fromResult["ok"]) {
                        chat = fromResult["result"];
                    }
                } catch (e) { }
                try {
                    var fromResult = await this.getUser(from["id"]);
                    if (fromResult["ok"]) {
                        from = fromResult["result"];
                    }
                } catch (e) { }
            }

            try {
                var get_message = await this.getMessage(chat["id"], update["message_id"], true, true);
                if (get_message["ok"]) {
                    if (get_message["result"]["update_message"] != null) {
                        json["message"] = get_message["result"]["update_message"];
                    }

                    if (get_message["result"]["update_channel_post"] != null) {
                        json["message"] = get_message["result"]["update_channel_post"];
                    }
                }
            } catch (e) { }


            json["message_id"] = update["message_id"];
            json["from"] = from;
            json["chat"] = chat;
            json["chat_instance"] = update["chat_instance"];
            json["data"] = Buffer.from(update["payload"]["data"], "base64").toString("utf-8");
            return {
                "ok": true,
                "result": { "callback_query": json }
            };
        }

        if (update["@type"] == "updateChatMember") {
            var json = {};
            var chat = { "id": update["chat_id"] };
            var from = { "id": update["actor_user_id"] };
            if (is_super_detail) {
                try {
                    var fromResult = await this.getChat(chat["id"]);
                    if (fromResult["ok"]) {
                        chat = fromResult["result"];
                    }
                } catch (e) { }
                try {
                    var fromResult = await this.getUser(from["id"]);
                    if (fromResult["ok"]) {
                        from = fromResult["result"];
                    }
                } catch (e) { }
            }
            json["from"] = from;
            json["chat"] = chat;
            json["date"] = update["date"];
            if (update["old_chat_member"]["@type"] == "chatMember") {
                var json_new_member = {};
                if (update["old_chat_member"]["member_id"]["@type"] == "messageSenderUser") {
                    var json_data_user = {
                        "id": update["old_chat_member"]["member_id"]["user_id"]
                    };
                    if (is_super_detail) {
                        try {
                            var fromResult = await this.getUser(json_data_user["id"]);
                            if (fromResult["ok"]) {
                                json_data_user = fromResult["result"];
                            }
                        } catch (e) { }
                    }
                    json_new_member["user"] = json_data_user;
                }
                json_new_member["status"] = String(update["old_chat_member"]["status"]["@type"]).toLocaleLowerCase().replace(RegExp("chatMemberStatus", "i"), "");
                json["old_member"] = json_new_member;
            }
            if (update["new_chat_member"]["@type"] == "chatMember") {
                var json_new_member = {};

                if (update["new_chat_member"]["member_id"]["@type"] == "messageSenderUser") {
                    var json_data_user = {
                        "id": update["new_chat_member"]["member_id"]["user_id"]
                    };
                    if (is_super_detail) {
                        try {
                            var fromResult = await this.getUser(json_data_user["id"]);
                            if (fromResult["ok"]) {
                                json_data_user = fromResult["result"];
                            }
                        } catch (e) { }
                    }
                    json_new_member["user"] = json_data_user;
                }

                json_new_member["status"] = String(update["new_chat_member"]["status"]["@type"]).toLocaleLowerCase().replace(RegExp("chatMemberStatus", "i"), "");
                json["new_member"] = json_new_member;
            }

            return {
                "ok": true,
                "result": { "chat_member": json }
            };
        }
        if (update["@type"] == "updateNewInlineQuery") {
            var json = {};
            var from = { "id": update["sender_user_id"] };
            json["id"] = update["id"];
            try {
                var fromResult = await this.getUser(from["id"]);
                if (fromResult["ok"]) {
                    from = fromResult["result"];
                }
            } catch (e) {
                console.log(e);
            }
            json["from"] = from;
            json["chat_type"] = String(update["chat_type"]["@type"]).toLocaleLowerCase().replace(RegExp("chatType", "i"), "");
            try {
                if (json["chat_type"] == "supergroup" && update["chat_type"]["is_channel"]) {
                    json["chat_type"] = "channel";
                }
            } catch (e) { }
            json["query"] = update["query"];
            json["offset"] = update["offset"];
            return {
                "ok": true,
                "result": { "inline_query": json }
            };
        }
        return json_result;
    }

    makeParametersApi(parameters = {}) {

        if (typeof parameters != "object") {
            parameters = {};
        }
        var jsonResult = { "@type": "" };

        try {
            var regexMethodSend = "^(sendMessage|sendPhoto|sendVideo|sendAudio|sendVoice|sendDocument|sendSticker|sendAnimation|editMessageText|copyMessage|forwardMessage)$";
            if (RegExp(regexMethodSend, "i").exec(parameters["@type"])) {
                jsonResult["@type"] = "sendMessage";
                if (RegExp("editMessage(Text)", "i").exec(parameters["@type"])) {
                    jsonResult["@type"] = parameters["@type"];
                }
                jsonResult["input_message_content"] = {
                    "@type": "inputMessageText",
                    "disableWebPagePreview": false,
                    "clearDraft": false
                };
                jsonResult["chat_id"] = parameters["chat_id"];
                if (typeof (parameters["disable_notification"]) == "boolean") {
                    jsonResult["disable_notification"] = parameters["disable_notification"];
                }
                if (typeof (parameters["reply_to_message_id"]) == "number") {
                    jsonResult["reply_to_message_id"] = parameters["reply_to_message_id"];
                }

                if (parameters["parse_mode"]) {
                    if (typeof (parameters["parse_mode"]) != "string") {
                        parameters["parse_mode"] = "html";
                    }
                } else {
                    parameters["parse_mode"] = "html";
                }

                if (typeof parameters["message_id"] == "number") {
                    jsonResult["message_id"] = parameters["message_id"];
                }
                if (typeof parameters["reply_markup"] == "object") {
                    jsonResult["reply_markup"] = this.reply_markup(parameters["reply_markup"]);
                }
                if (RegExp("^(sendMessage|editMessageText)$", "i").exec(parameters["@type"])) {
                    var text = this.parseMode(
                        parameters["text"],
                        parameters["parse_mode"],
                        parameters["entities"],
                    );
                    jsonResult["input_message_content"]["@type"] = "inputMessageText";
                    jsonResult["input_message_content"]["text"] = text;
                }
                if (RegExp("^(sendPhoto)$", "i").exec(parameters["@type"])) {
                    var getDetailFile = this.typeFile(parameters["photo"]);
                    jsonResult["input_message_content"]["@type"] = "inputMessagePhoto";
                    jsonResult["input_message_content"]["photo"] = getDetailFile;
                }
                if (RegExp("^(sendVoice)$", "i").exec(parameters["@type"])) {
                    var getDetailFile = this.typeFile(parameters["voice"]);
                    jsonResult["input_message_content"]["@type"] = "inputMessageVoiceNote";
                    jsonResult["input_message_content"]["voice_note"] = getDetailFile;
                }
                if (RegExp("^(sendSticker)$", "i").exec(parameters["@type"])) {
                    var getDetailFile = this.typeFile(parameters["sticker"]);
                    jsonResult["input_message_content"]["@type"] = "inputMessageSticker";
                    jsonResult["input_message_content"]["sticker"] = getDetailFile;
                }
                if (RegExp("^(sendAnimation)$", "i").exec(parameters["@type"])) {
                    var getDetailFile = this.typeFile(parameters["animation"]);
                    jsonResult["input_message_content"]["@type"] = "inputMessageAnimation";
                    jsonResult["input_message_content"]["animation"] = getDetailFile;
                }
                if (RegExp("^(sendDocument)$", "i").exec(parameters["@type"])) {
                    var getDetailFile = this.typeFile(parameters["document"]);
                    jsonResult["input_message_content"]["@type"] = "inputMessageDocument";
                    jsonResult["input_message_content"]["document"] = getDetailFile;
                }
                if (RegExp("^(sendAudio)$", "i").exec(parameters["@type"])) {
                    var getDetailFile = this.typeFile(parameters["audio"]);
                    jsonResult["input_message_content"]["@type"] = "inputMessageAudio";
                    jsonResult["input_message_content"]["audio"] = getDetailFile;
                }
                if (RegExp("^(sendVideo)$", "i").exec(parameters["@type"])) {
                    var getDetailFile = this.typeFile(parameters["video"]);
                    jsonResult["input_message_content"]["@type"] = "inputMessageVideo";
                    jsonResult["input_message_content"]["video"] = getDetailFile;
                }
                if (RegExp("^((copy|forward)Message)$", "i").exec(parameters["@type"])) {
                    jsonResult["input_message_content"]["@type"] = "inputMessageForwarded";
                    jsonResult["input_message_content"]["from_chat_id"] = parameters["from_chat_id"];
                    jsonResult["input_message_content"]["message_id"] = parameters["message_id"];
                    jsonResult["input_message_content"]["copy_options"] = {
                        '@type': "messageCopyOptions",
                        "send_copy": RegExp("copyMessage", "i").test(parameters["@type"])
                    };
                }
                if (!RegExp("^(sendMessage|sendLocation|sendSticker)$", "i").exec(parameters["@type"])) {
                    if (parameters["caption"]) {
                        var caption = this.parseMode(
                            parameters["caption"],
                            parameters["parse_mode"],
                            parameters["entities"],
                        );
                        jsonResult["input_message_content"]["caption"] = caption;
                    }
                }

                return jsonResult;
            }

            if (RegExp("^answerInlineQuery$", "i").exec(parameters["@type"])) {
                parameters["@type"] = "answerInlineQuery";

                if (typeof (parameters["results"]) == "array") {
                    var array_results = [];
                    for (var i = 0; i < parameters["results"].length; i++) {
                        var loop_data = parameters["results"][i];

                        if (typeof (loop_data["type"]) == "string") {
                            loop_data["@type"] = loop_data["type"];
                            delete loop_data["type"];
                        }
                        if (typeof (loop_data["id"]) != "string") {
                            loop_data["id"] ??= "$i";
                            loop_data["id"] = (String(loop_data["id"]));
                        }

                        if (typeof (loop_data["reply_markup"]) == "object") {
                            loop_data["reply_markup"] = (this.reply_markup(loop_data["reply_markup"]));
                        }
                        array_results.push(loop_data);
                    }
                    parameters["results"] = array_results;
                }
                return parameters;
            }

            return parameters;

        } catch (e) {
            console.log(e);
            return parameters;
        }
        return parameters;
    }

    typeFile(content) {
        var data = {};
        if (RegExp("^http", "i").exec(content)) {
            data = { "@type": 'inputFileRemote', "id": content };
        } else if (RegExp("^(\/|\.\.?\/|~\/)", "i").exec(content)) {
            data = { "@type": 'inputFileLocal', "path": content };
        } else if (typeof (content) == 'number') {
            data = { "@type": 'inputFileId', "id": content };
        } else {
            data = { "@type": 'inputFileRemote', "id": content };
        }
        return data;
    }
    reply_markup(keyboard) {
        try {
            if (keyboard["inline_keyboard"].length > 0) {
                var json = { "@type": "replyMarkupInlineKeyboard" }
                var array_rows = []
                for (var i = 0; i < keyboard["inline_keyboard"].length; i++) {
                    var loop_array_keyboard = keyboard["inline_keyboard"][i]
                    var array_loop = []
                    for (var ii = 0; ii < loop_array_keyboard.length; ii++) {
                        var in_loop_array_keyboard = loop_array_keyboard[ii]
                        var in_json_keyboard = { "@type": "inlineKeyboardButton" }
                        if (in_loop_array_keyboard["text"]) {
                            in_json_keyboard["text"] = in_loop_array_keyboard["text"];
                        }

                        if (in_loop_array_keyboard["url"]) {
                            in_json_keyboard["type"] = {
                                "@type": "inlineKeyboardButtonTypeUrl",
                                "url": in_loop_array_keyboard["url"]
                            };
                        }

                        if (in_loop_array_keyboard["callback_data"]) {
                            in_json_keyboard["type"] = {
                                "@type": "inlineKeyboardButtonTypeCallback",
                                "data": Buffer.from(in_loop_array_keyboard["callback_data"]).toString('base64')
                            };
                        }
                        array_loop.push(in_json_keyboard);
                    }
                    array_rows.push(array_loop);
                }
                json.rows = array_rows;
                return json;
            }

        } catch (e) {
            console.log(e);
            return null;
        }
    }


    json(query) {
        return Buffer.from(JSON.stringify(query) + '\0');
    }

    on(type, callback) {
        this.event_emitter.on(type, async function (update) {
            return callback(update);
        });
    }

    async getChat(chat_id, is_api_update = false) {
        var json_result = {
            "ok": true,
            "result": {
                "id": chat_id,
                "first_name": "",
                "last_name": "",
                "title": "",
                "type": "",
                "detail": {}
            }
        };
        try {
            if (typeof chat_id == "string" && RegExp("@?[a-z0-9_]+", "i").exec(chat_id)) {
                var search_public_chat = await this.invoke("searchPublicChat", { "username": chat_id });
                if (search_public_chat["@type"] == "chat") {
                    chat_id = search_public_chat["id"];
                }
            }
            var get_chat = await this.invoke("getChat", { "chat_id": chat_id });
            var json = {};
            if (get_chat["@type"] == "chat") {
                var type_chat = String(get_chat["type"]["@type"]).toLocaleLowerCase().replace(RegExp("chatType", "i"), "");
                if (type_chat == "supergroup") {
                    var getSupergroup = await this.invoke("getSupergroup", {
                        "supergroup_id": Number(String(chat_id).replace(RegExp("^-100", "i"), ""))
                    });
                    json["id"] = chat_id;
                    json["title"] = get_chat["title"];
                    if (getSupergroup["username"]) {
                        json["username"] = getSupergroup["username"];
                    }
                    if (typeof (getSupergroup["status"]) == "object") {
                        json["status"] = String(getSupergroup["status"]["@type"]).toLocaleLowerCase().replace(RegExp("chatMemberStatus", "i"), "");
                    }
                    json["type"] = get_chat["type"]["is_channel"] ? "channel" : "supergroup";
                    json["detail"] = {
                        "member_count": getSupergroup["member_count"],
                        "administrator_count": 0,
                        "restricted_count": 0,
                        "banned_count": 0,
                        "has_protected_content": get_chat["has_protected_content"] ?? false,
                        "is_marked_as_unread": get_chat["is_marked_as_unread"] ?? false,
                        "is_blocked": get_chat["is_blocked"] ?? false,
                        "has_scheduled_messages": get_chat["has_scheduled_messages"] ?? false,
                        "can_be_deleted_only_for_self": get_chat["can_be_deleted_only_for_self"] ?? false,
                        "can_be_deleted_for_all_users": get_chat["can_be_deleted_for_all_users"] ?? false,
                        "can_be_reported": get_chat["can_be_reported"] ?? false,
                        "default_disable_notification": get_chat["default_disable_notification"] ?? false,
                        "unread_count": get_chat["unread_count"] ?? 0,
                        "last_read_inbox_message_id": get_chat["last_read_inbox_message_id"] ?? 0,
                        "last_read_outbox_message_id": get_chat["last_read_outbox_message_id"] ?? 0,
                        "unread_mention_count": get_chat["unread_mention_count"] ?? 0,
                        "has_linked_chat": getSupergroup["has_linked_chat"],
                        "has_location": getSupergroup["has_location"],
                        "sign_messages": getSupergroup["sign_messages"],
                        "is_slow_mode_enabled": getSupergroup["is_slow_mode_enabled"],
                        "is_broadcast_group": getSupergroup["is_broadcast_group"],
                        "is_verified": getSupergroup["is_verified"],
                        "is_scam": getSupergroup["is_scam"],
                        "is_fake": getSupergroup["is_fake"]
                    };
                    return { "ok": true, "result": json };
                } else if (type_chat == "basicgroup") {
                    var getBasicGroup = await this.invoke("getBasicGroup", {
                        "basic_group_id": Number(String(chat_id).replace(RegExp("^-", "i"), ""))
                    });
                    json["id"] = chat_id;
                    json["title"] = get_chat["title"];
                    if (typeof (getBasicGroup["status"]) == "object") {
                        json["status"] = String(getBasicGroup["status"]["@type"]).toLocaleLowerCase().replace(RegExp("chatMemberStatus", "i"), "");
                    }
                    json["type"] = "group";
                    json["detail"] = {
                        "member_count": getBasicGroup["member_count"],
                        "has_protected_content": get_chat["has_protected_content"] ?? false,
                        "is_marked_as_unread": get_chat["is_marked_as_unread"] ?? false,
                        "is_blocked": get_chat["is_blocked"] ?? false,
                        "has_scheduled_messages": get_chat["has_scheduled_messages"] ?? false,
                        "can_be_deleted_only_for_self": get_chat["can_be_deleted_only_for_self"] ?? false,
                        "can_be_deleted_for_all_users": get_chat["can_be_deleted_for_all_users"] ?? false,
                        "can_be_reported": get_chat["can_be_reported"] ?? false,
                        "default_disable_notification": get_chat["default_disable_notification"] ?? false,
                        "unread_count": get_chat["unread_count"] ?? 0,
                        "last_read_inbox_message_id": get_chat["last_read_inbox_message_id"] ?? 0,
                        "last_read_outbox_message_id": get_chat["last_read_outbox_message_id"] ?? 0,
                        "unread_mention_count": get_chat["unread_mention_count"] ?? 0,
                    };
                    return { "ok": true, "result": json };
                } else if (type_chat == "private") {
                    var get_user = await this.invoke("getUser", { "user_id": chat_id, });
                    if (RegExp("^user$", "i").exec(get_user["@type"])) {
                        var json = {};
                        json["id"] = get_user["id"];
                        try {
                            if (RegExp("^userTypeBot$", "i").exec(get_user["type"]["@type"])) {
                                json["is_bot"] = true;
                            } else {
                                json["is_bot"] = false;
                            }
                        } catch (e) {
                            json["is_bot"] = false;
                        }
                        json["first_name"] = get_user["first_name"];
                        if (get_user["last_name"]) {
                            json["last_name"] = get_user["last_name"];
                        }
                        if (get_user["username"]) {
                            json["username"] = get_user["username"];
                        }
                        if (get_user["phone_number"]) {
                            json["phone_number"] = get_user["phone_number"];
                        }
                        if (get_user["language_code"]) {
                            json["language_code"] = get_user["language_code"];
                        }
                        json["type"] = 'private';
                        json["detail"] = {
                            "has_protected_content": get_chat["has_protected_content"] ?? false,
                            "is_marked_as_unread": get_chat["is_marked_as_unread"] ?? false,
                            "is_blocked": get_chat["is_blocked"] ?? false,
                            "has_scheduled_messages": get_chat["has_scheduled_messages"] ?? false,
                            "can_be_deleted_only_for_self": get_chat["can_be_deleted_only_for_self"] ?? false,
                            "can_be_deleted_for_all_users": get_chat["can_be_deleted_for_all_users"] ?? false,
                            "can_be_reported": get_chat["can_be_reported"] ?? false,
                            "default_disable_notification": get_chat["default_disable_notification"] ?? false,
                            "unread_count": get_chat["unread_count"] ?? 0,
                            "last_read_inbox_message_id": get_chat["last_read_inbox_message_id"] ?? 0,
                            "last_read_outbox_message_id": get_chat["last_read_outbox_message_id"] ?? 0,
                            "unread_mention_count": get_chat["unread_mention_count"] ?? 0,
                            "is_contact": get_user["is_contact"],
                            "is_mutual_contact": get_user["is_mutual_contact"],
                            "is_verified": get_user["is_verified"],
                            "is_support": get_user["is_support"],
                            "is_scam": get_user["is_scam"],
                            "is_fake": get_user["is_fake"],
                            "have_acces": get_user["have_access"]
                        };
                        return { "ok": true, "result": json };
                    }

                }
            }
        } catch (e) {
            if (is_api_update) {
                return json_result;
            } else {
                throw e;
            }
        }
        return json_result;
    }
    async getUser(chat_id, is_api_update = false) {
        var json_result = {
            "ok": true,
            "result": {
                "id": chat_id,
                "first_name": "",
                "last_name": "",
                "title": "",
                "type": "",
                "detail": {}
            }
        };
        try {
            if (typeof chat_id == "string" && RegExp("@?[a-z0-9_]+", "i").exec(chat_id)) {
                var search_public_chat = await this.invoke("searchPublicChat", { "username": chat_id });
                if (search_public_chat["@type"] == "chat") {
                    chat_id = search_public_chat["id"];
                }
            }
            var get_user = await this.invoke("getUser", { "user_id": chat_id, });
            if (RegExp("^user$", "i").exec(get_user["@type"])) {
                var json = {};
                json["id"] = get_user["id"];
                try {
                    if (RegExp("^userTypeBot$", "i").exec(get_user["type"]["@type"])) {
                        json["is_bot"] = true;
                    } else {
                        json["is_bot"] = false;
                    }
                } catch (e) {
                    json["is_bot"] = false;
                }
                json["first_name"] = get_user["first_name"];
                if (get_user["last_name"]) {
                    json["last_name"] = get_user["last_name"];
                }
                if (get_user["username"]) {
                    json["username"] = get_user["username"];
                }
                if (get_user["phone_number"]) {
                    json["phone_number"] = get_user["phone_number"];
                }
                if (get_user["language_code"]) {
                    json["language_code"] = get_user["language_code"];
                }
                json["type"] = 'private';
                json["detail"] = {
                    "is_contact": get_user["is_contact"],
                    "is_mutual_contact": get_user["is_mutual_contact"],
                    "is_verified": get_user["is_verified"],
                    "is_support": get_user["is_support"],
                    "is_scam": get_user["is_scam"],
                    "is_fake": get_user["is_fake"],
                    "have_acces": get_user["have_access"]
                };
                return { "ok": true, "result": json };
            }



        } catch (e) {
            if (is_api_update) {
                return json_result;
            } else {
                throw e;
            }
        }
        return json_result;
    }
    parseMode(text, parse_mode, entities) {
        var pesan = { "text": text };
        var parseMode = 'textParseModeHTML'
        if (typeof parse_mode == "string") {
            parse_mode = String(parse_mode).toLocaleLowerCase();
            if (parse_mode == 'markdown') {
                parseMode = 'textParseModeMarkdown'
            } else if (parse_mode == 'html') {
                parseMode = 'textParseModeHTML';
            } else {
                parse_mode = false;
            }
        }

        if (typeof parse_mode == "string") {
            pesan = this.client_execute({
                "@type": 'parseTextEntities',
                "parse_mode": { "@type": parseMode },
                "text": text
            });
        }

        if (typeof entities == "object") {
            pesan = {
                "@type": 'formattedText',
                "text": text,
                "entities": entities
            };
        }
        return pesan;
    }

    async initIsolate() {
        while (true) {
            await timer.setTimeout(1);
            var update = this.client_receive();
            if (update != null) {
                if (update["@type"] == "updateAuthorizationState") {
                    var authState = update["authorization_state"];
                    if (authState["@type"] == "authorizationStateWaitTdlibParameters") {
                        this.client_send(this.optionTdlibDefault, true);
                    }
                    if (authState["@type"] == "authorizationStateWaitEncryptionKey") {
                        var isEncrypted = authState['is_encrypted'] ?? false;
                        if (isEncrypted) {
                            this.client_send({
                                "@type": 'checkDatabaseEncryptionKey',
                                'encryption_key': this.optionTdlibDefault["database_key"]
                            }, true);
                        } else {
                            this.client_send({
                                '@type': 'setDatabaseEncryptionKey',
                                'new_encryption_key': this.optionTdlibDefault["database_key"]
                            }, true);
                        }

                    }
                }
                this.event_emitter.emit("update", update);
            }
        }
    }

    async getMe() {
        var get_me = await this.invoke("getMe");
        var result = {};
        result["id"] = get_me["id"];
        result["is_bot"] = false;
        result["first_name"] = get_me["first_name"];
        if (get_me["last_name"]) {
            result["last_name"] = get_me["last_name"];
        }
        if (get_me["username"]) {
            result["username"] = get_me["username"];
        }
        if (get_me["phone_number"]) {
            result["phone_number"] = get_me["phone_number"];
        }
        result["status"] = String(get_me["status"]["@type"]).toLocaleLowerCase().replace(RegExp("userStatus", "i"), "");
        result["type_account"] = String(get_me["type"]["@type"]).toLocaleLowerCase().replace(RegExp("userType", "i"), "");
        result["type"] = "private";
        if (result["type_account"] == "bot") {
            result["is_bot"] = true;
        }
        if (get_me["language_code"].toString().isNotEmpty) {
            result["language_code"] = get_me["language_code"];
        }
        result["detail"] = {};
        for (var key in get_me) {
            if (Object.hasOwnProperty.call(get_me, key)) {
                var value = get_me[key];
                if (typeof (value) == "boolean") {
                    result["detail"][String(key)] = value;
                }

            }
        }
        for (var key in get_me["type"]) {
            if (Object.hasOwnProperty.call(get_me["type"], key)) {
                var value = get_me["type"][key];
                if (typeof (value) == "boolean") {
                    result["detail"][key.toString()] = value;
                }
            }
        }
        return { "ok": true, "result": result };
    }


    async raw_api_light(update) {
        if (RegExp("updateNewMessage|updateChatMember|updateNewCallbackQuery|updateNewInlineQuery", "i").exec(update["@type"])) {
            try {
                var getMessage = await this.jsonMessage((update["@type"] == "updateNewMessage") ? update["message"] : update, true, false);
                if (getMessage["ok"]) {
                    return getMessage["result"];
                }
                return update;
            } catch (e) {
                console.log(e);
                return update;
            }
        } else {
            return update;
        }
    }
    async raw_api(update) {
        if (RegExp("updateNewMessage|updateChatMember|updateNewCallbackQuery|updateNewInlineQuery", "i").exec(update["@type"])) {
            try {
                var getMessage = await this.jsonMessage((update["@type"] == "updateNewMessage") ? update["message"] : update, true, true);
                if (getMessage["ok"]) {
                    return getMessage["result"];
                }
                return update;
            } catch (e) {
                console.log(e);
                return update;
            }
        } else {
            return update;
        }
    }

}



module.exports = {
    Tdlib
};
